module particle_ec_calc

  use vtype_module

  implicit none

  private

  public :: ParticleECCalc

  logical, parameter :: FlagChkRangeAndStop = .true.
!!$  logical, parameter :: FlagChkRangeAndStop = .false.

  integer, parameter :: IDSizeDistFuncMarshallPalmer = 11
  integer, parameter :: IDSizeDistFuncGamma          = 12
  integer, parameter :: IDSizeDistFuncModifiedGamma  = 13
  integer, parameter :: IDSizeDistFuncLogNormal      = 14


  interface ParticleECCalc
!!$    module procedure ParticleECCalcSerial
    module procedure ParticleECCalcOMP
  end interface ParticleECCalc


  character(stdstr), save :: ModuleName = 'particle_ec_calc'


  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------

  function SizeDistMarshallPalmer( Slope, Radius ) result( SDF )

    real(dp), intent(in ) :: Slope
    real(dp), intent(in ) :: Radius

    real(dp) :: SDF

    ! Local variables
    !
    SDF = exp( - Slope * Radius )


  end function SizeDistMarshallPalmer

  !----------------------------------------------------------------------------

  function SizeDistGamma( Mu, Slope, Radius ) result( SDF )

    real(dp), intent(in ) :: Mu
    real(dp), intent(in ) :: Slope
    real(dp), intent(in ) :: Radius

    real(dp) :: SDF

    ! Local variables
    !
    SDF = Radius**Mu * exp( - Slope * Radius )


  end function SizeDistGamma

  !----------------------------------------------------------------------------
!!$
!!$  function SizeDistModifiedGamma( RadiusEff, VarEff, Radius ) result( SDF )
!!$
!!$    real(dp), intent(in ) :: RadiusEff
!!$    real(dp), intent(in ) :: VarEff
!!$    real(dp), intent(in ) :: Radius
!!$
!!$    real(dp) :: SDF
!!$
!!$    ! Local variables
!!$    !
!!$
!!$    ! Eq. (13) of Petty and Huang (2011)
!!$    SDF = Radius**( ( 1.0d0 - 3.0d0 * VarEff ) / VarEff ) &
!!$      & * exp( - Radius / ( RadiusEff * VarEff ) )
!!$
!!$
!!$  end function SizeDistModifiedGamma
!!$
  !----------------------------------------------------------------------------

  function SizeDistLogNormal( MedianRadius, lnS, Radius ) result( SDF )

    real(dp), intent(in ) :: MedianRadius
    real(dp), intent(in ) :: lnS
    real(dp), intent(in ) :: Radius

    real(dp) :: SDF

    ! Local variables
    !

    SDF = 1.0d0 / ( lnS * Radius ) &
      & * exp( - ( log( Radius / MedianRadius ) / lnS )**2 / 2.0d0 )


  end function SizeDistLogNormal

  !----------------------------------------------------------------------------

  subroutine CalcParamLogNormal( &
    & EffRadius, StdDev, &
    & MedianRadius, ModeRadius, lnS  &
    & )

    real(dp), intent(in ) :: EffRadius
    real(dp), intent(in ) :: StdDev
    real(dp), intent(out) :: MedianRadius
    real(dp), intent(out) :: ModeRadius
    real(dp), intent(out) :: lnS

    ! Local variables
    !
!!$    real(DP) :: X0
    real(DP) :: X1
    real(DP) :: X2
    real(DP) :: X
    real(DP) :: FX1
    real(DP) :: FX2
    real(DP) :: FX
!!$    real(DP) :: FDerivX
    integer  :: iLoop
    integer, parameter :: NLoop = 100
    real(DP) :: a_X(NLoop)


    ! a line which crosses ( x0, f(x0) )
    !   y = f'(x0) * ( x - x0 ) + f(x0)
    ! x value when y = 0
    !   0 = f'(x0) * ( x - x0 ) + f(x0)
    !   f'(x0) * ( x - x0 ) = - f(x0)
    !   x - x0 = - f(x0) / f'(x0)
    !   x = x0 - f(x0) / f'(x0)

    ! f'(x) = 6/5*re**(4/5)*x**(1/5) - 8/5*re**(2/5)*x**(3/5)
    ! f'(x) = 0
    ! 6/5*re**(4/5)*x**(1/5) - 8/5*re**(2/5)*x**(3/5) = 0
    ! 6*re**(4/5)*x**(1/5) = 8*re**(2/5)*x**(3/5)
    ! 6**5*re**4*x**1 = 8**5*re**2*x**3
    ! 6**5*re**2 = 8**5*x**2
    ! x**2 = (6/8)**5*re**2
    ! x = (6/8)**(5/2) * re
    ! x = (3/4)**(5/2) * re


    ! Newton method
    !
!!$    X  = EffRadius
!!$    X0 = X
!!$    do iLoop = 1, NLoop
!!$      FX = &
!!$        &   EffRadius**(4.0d0/5.0d0) * X**(6.0d0/5.0d0) &
!!$        & - EffRadius**(2.0d0/5.0d0) * X**(8.0d0/5.0d0) &
!!$        & - StdDev**2
!!$      FDerivX = &
!!$        &   6.0d0/5.0d0 * EffRadius**(4.0d0/5.0d0) * X**(1.0d0/5.0d0) &
!!$        & - 8.0d0/5.0d0 * EffRadius**(2.0d0/5.0d0) * X**(3.0d0/5.0d0)
!!$      X = X - FX / FDerivX
!!$      X = max( X, (3.0d0/4.0d0)**(5.0d0/2.0d0) * EffRadius )
!!$
!!$      write( 6, * ) EffRadius, StdDev, X0, X, FX, FDerivX
!!$
!!$      if ( ( abs( X - X0 ) / X0 ) < 1.0d-5 ) exit
!!$      X0 = X
!!$      a_X(iLoop) = X
!!$    end do
!!$    if ( iLoop > NLoop ) then
!!$      write( 6, * ) "Newton method does not converge."
!!$      do iLoop = 1, NLoop
!!$        write( 60, * ) iLoop, a_X(iLoop)
!!$      end do
!!$      stop
!!$    end if


    ! Bisection method
    !
    X1 = (3.0d0/4.0d0)**(5.0d0/2.0d0) * EffRadius
    X2 = EffRadius * 2.0d0
    do iLoop = 1, NLoop
      X = ( X1 + X2 ) / 2.0d0

      FX1 = &
        &   EffRadius**(4.0d0/5.0d0) * X1**(6.0d0/5.0d0) &
        & - EffRadius**(2.0d0/5.0d0) * X1**(8.0d0/5.0d0) &
        & - StdDev**2
      FX2 = &
        &   EffRadius**(4.0d0/5.0d0) * X2**(6.0d0/5.0d0) &
        & - EffRadius**(2.0d0/5.0d0) * X2**(8.0d0/5.0d0) &
        & - StdDev**2
      FX  = &
        &   EffRadius**(4.0d0/5.0d0) * X**(6.0d0/5.0d0)  &
        & - EffRadius**(2.0d0/5.0d0) * X**(8.0d0/5.0d0)  &
        & - StdDev**2

      if ( FX1*FX <= 0.0d0 ) then
        X2 = X
      else if ( FX*FX2 < 0.0d0 ) then
        X1 = X
      else
        write( 6, * ) 'Zero point is not included in a range from X1 to X2.'
        write( 6, * ) X1,FX1
        write( 6, * ) X ,FX
        write( 6, * ) X2,FX2
        write( 6, * ) EffRadius, StdDev
        stop
      end if


      if ( ( abs( X2 - X1 ) / X2 ) < 1.0d-5 ) exit
      a_X(iLoop) = X
    end do
    if ( iLoop > NLoop ) then
      write( 6, * ) "Bisection method does not converge."
      do iLoop = 1, NLoop
        write( 60, * ) iLoop, a_X(iLoop)
      end do
      stop
    end if


    MedianRadius = X
    if ( EffRadius >= MedianRadius ) then
      ! exp( 5.0d0/2.0d0 * lnS**2 ) = EffRadius/MedianRadius
      ! 5.0d0/2.0d0 * lnS**2 = log( EffRadius/MedianRadius )
      ! lnS**2 = 2.0d0/5.0d0 * log( EffRadius/MedianRadius )
      lnS = sqrt( 2.0d0/5.0d0 * log( EffRadius/MedianRadius ) )
    else
      lnS = 0.0d0
    end if
    ModeRadius   = exp( log( MedianRadius ) - lnS**2 )


  end subroutine CalcParamLogNormal

  !----------------------------------------------------------------------------

  subroutine SetRadiusIntegRangeLogNormal( &
    & lnEffRadius, lnMedianRadius, lnS,    & ! (in)
    & lnRadiusS, lnRadiusE                 & ! (out)
    & )

    real(DP), intent(in ) :: lnEffRadius
    real(DP), intent(in ) :: lnMedianRadius
    real(DP), intent(in ) :: lnS
    real(DP), intent(out) :: lnRadiusS
    real(DP), intent(out) :: lnRadiusE


    ! Local variables
    !
    real(DP) :: NormWidth

    NormWidth = 4.0d0


    if ( lnS == 0 ) then
      lnRadiusS = lnEffRadius
      lnRadiusE = lnRadiusS
    else
      lnRadiusS = lnMedianRadius - NormWidth * lnS
      lnRadiusE = lnMedianRadius + NormWidth * lnS
    end if


  end subroutine SetRadiusIntegRangeLogNormal

  !----------------------------------------------------------------------------

  subroutine InterpolateParams( &
    & NRefRadius, r_RefRadius, r_RefQExt, r_RefSSA, r_RefAF, &
    & Radius, &
    & QExt, SSA, AF &
    & )

    integer , intent(in ) :: NRefRadius
    real(dp), intent(in ) :: r_RefRadius(NRefRadius)
    real(dp), intent(in ) :: r_RefQExt  (NRefRadius)
    real(dp), intent(in ) :: r_RefSSA   (NRefRadius)
    real(dp), intent(in ) :: r_RefAF    (NRefRadius)
    real(dp), intent(in ) :: Radius
    real(dp), intent(out) :: QExt
    real(dp), intent(out) :: SSA
    real(dp), intent(out) :: AF


    ! Local variables
    !
    integer :: iRadius


    if ( Radius < r_RefRadius(1) ) then
      write( 6, * ) 'Unexpected error: too small radius, ', Radius
      stop
      QExt = -1.0d0
      SSA  = -1.0d0
      AF   = -1.0d0
    else if ( r_RefRadius(NRefRadius) < Radius ) then
      write( 6, * ) 'Unexpected error: too large radius, ', Radius
      stop
      QExt = -1.0d0
      SSA  = -1.0d0
      AF   = -1.0d0
    else

      if ( NRefRadius == 1 ) then
        iRadius = 1
        QExt = r_RefQExt(iRadius)
        SSA  = r_RefSSA (iRadius)
        AF   = r_RefAF  (iRadius)
      else
        do iRadius = 1+1, NRefRadius
          if ( Radius <= r_RefRadius(iRadius) ) exit
        end do
        if ( NRefRadius < iRadius ) then
          write( 6, * ) 'In InterpolateParams : Unexpected error'
          write( 6, * ) '  Unable to find an appropriate index for interpolation'
          write( 6, * ) r_RefRadius(NRefRadius), Radius
          stop
        end if
        QExt = &
          &   ( r_RefQExt  (iRadius) - r_RefQExt  (iRadius-1) ) &
          & / ( r_RefRadius(iRadius) - r_RefRadius(iRadius-1) ) &
          & * ( Radius               - r_RefRadius(iRadius-1) ) &
          & + r_RefQExt(iRadius-1)
        SSA = &
          &   ( r_RefSSA   (iRadius) - r_RefSSA   (iRadius-1) ) &
          & / ( r_RefRadius(iRadius) - r_RefRadius(iRadius-1) ) &
          & * ( Radius               - r_RefRadius(iRadius-1) ) &
          & + r_RefSSA (iRadius-1)
        AF = &
          &   ( r_RefAF    (iRadius) - r_RefAF    (iRadius-1) ) &
          & / ( r_RefRadius(iRadius) - r_RefRadius(iRadius-1) ) &
          & * ( Radius               - r_RefRadius(iRadius-1) ) &
          & + r_RefAF  (iRadius-1)
      end if

    end if


  end subroutine InterpolateParams

  !----------------------------------------------------------------------------

  ! NOT USED, 2019/10/30

  subroutine ParticleECCalcSerial( &
    & DelWaveNum, WaveNumS, WaveNumE, &
    & MolNumNonRadAct, &
    & NMolNum, m_MolNum, &
    & NParticle, &
    & a_FlagReadQExtSSAAFFromFile, &
    & a_FlagReadQExtSSAAFFromParam, &
    & a_ParticleName, a_ParticleSDF, a_ParticleDataFN, &
    & a_ParticleNEffRadius, a_ParticleEffRadiusS, a_ParticleEffRadiusE, &
    & a_ParticleStdDev, &
    & a_ParticleNRadiusInteg, a_ParticleRadiusIntegWidth, &
    & OutputFileName &
    & )

    use constants0, only: &
      & PI

    use gauss_quad, only : GauLeg

    use rayscat, only : CalcRayScatCoef

    use refindex, only : &
      & RefIndexInit, &
      & RefIndexCalc, &
      & RefIndexFinalize

    use mie_theory, only : CALLBH3

    use particle_ec_io_module, only : &
      & particle_ec_io_output_init, &
      & particle_ec_io_output_end, &
      & particle_ec_io_output

    use h2o_water, only : &
      & H2OWaterInit, &
      & H2OWaterFinalize, &
      & H2OWaterCalcParam
    use h2o_ice, only : &
      & H2OIceInit, &
      & H2OIceFinalize, &
      & H2OIceCalcParam


    real(dp)    , intent(in ) :: DelWaveNum
    real(dp)    , intent(in ) :: WaveNumS
    real(dp)    , intent(in ) :: WaveNumE
    integer     , intent(in ) :: MolNumNonRadAct
    integer     , intent(in ) :: NMolNum
    integer     , intent(in ) :: m_MolNum(NMolNum)
    integer     , intent(in ) :: NParticle
    logical     , intent(in ) :: a_FlagReadQExtSSAAFFromFile (NParticle)
    logical     , intent(in ) :: a_FlagReadQExtSSAAFFromParam(NParticle)
    character(*), intent(in ) :: a_ParticleName            (NParticle)
    character(*), intent(in ) :: a_ParticleSDF             (NParticle)
    character(*), intent(in ) :: a_ParticleDataFN          (NParticle)
    integer     , intent(in ) :: a_ParticleNEffRadius      (NParticle)
    real(dp)    , intent(in ) :: a_ParticleEffRadiusS      (NParticle)
    real(dp)    , intent(in ) :: a_ParticleEffRadiusE      (NParticle)
    real(dp)    , intent(in ) :: a_ParticleStdDev          (NParticle)
    integer     , intent(in ) :: a_ParticleNRadiusInteg    (NParticle)
    real(dp)    , intent(in ) :: a_ParticleRadiusIntegWidth(NParticle)
    character(*), intent(in ) :: OutputFileName


    ! a_ParticleName             : Particle name
    ! a_ParticleSDF              : Particle size distribution function
    ! a_ParticleDataFN           : Data filename for optical properties
    ! a_ParticleNEffRadius       : Number of effective radii
    ! a_ParticleEffRadiusS       : Start effective radius in a table
    ! a_ParticleEffRadiusE       : End   effective radius in a table
    ! a_ParticleStdDev           : Standard deviation of size distribution
    !                            : Valid only for lognormal distribution
    ! a_ParticleNRadiusInteg     : Number of points for radius integration
    ! a_ParticleRadiusIntegWidth : Radius width for radius integration

!!$    allocate( z_Press ( NLev    ) )
!!$    allocate( m_MolNum( NMolNum ) )
!!$    allocate( z_Temp  ( NLev    ) )
!!$    allocate( zm_VMR  ( NLev, NMolNum ) )


    ! Local variables
    !
    real(dp)    , parameter   :: MissingValue = -1.0d100

    integer(i4b)              :: NWaveNum
    real(dp)    , allocatable :: w_WaveNum   (:)
    real(dp)    , allocatable :: w_Wavelength(:)

    real(dp)                  :: DelWaveNumMod


    real(dp)    , allocatable :: w_RayScatCoefNonRadAct(:)
    real(dp)    , allocatable :: wm_RayScatCoef(:,:)
    real(dp)    , allocatable :: wra_QExt (:,:,:)
    real(dp)    , allocatable :: wra_SSA  (:,:,:)
    real(dp)    , allocatable :: wra_AF   (:,:,:)

    integer                   :: NRadiusInteg
    integer                   :: iRadius
    real(DP)                  :: EffRadius
    real(DP)                  :: RadiusIntegWidth
    real(DP)                  :: QExt
    real(DP)                  :: SSA
    real(DP)                  :: SSCoA
    real(DP)                  :: AF

!!$    real(DP)                  :: a_ParticleRadiusStd(NParticle)
    real(DP)                  :: Sum1
    real(DP)                  :: Sum2
    real(DP)                  :: Sum3
    real(DP)                  :: SDF

    integer            :: IDSizeDistFunc
    integer, parameter :: IDSizeDistFuncMarshallPalmer = 11
    integer, parameter :: IDSizeDistFuncGamma          = 12
    integer, parameter :: IDSizeDistFuncModifiedGamma  = 13
    integer, parameter :: IDSizeDistFuncLogNormal      = 14

    ! variables for size distribution
    real(DP)                  :: Slope
    ! variables for Gamma distribution
    real(DP)                  :: Mu
    !
    real(DP)                  :: MedianRadius
    real(DP)                  :: ModeRadius
    real(DP)                  :: lnS

!!$    real(dp)    , parameter   :: Rmed = 1.00045D0
    real(dp)    , parameter   :: Rmed = 1.0d0

    real(DP)                  :: RIReal
    real(DP)                  :: RIImag

    integer(i4b)              :: NRefRadius
    real(DP), allocatable     :: r_RefRadius(:)
    real(DP), allocatable     :: r_RefQExt  (:)
    real(DP), allocatable     :: r_RefSSA   (:)
    real(DP), allocatable     :: r_RefAF    (:)

    real(DP)                  :: RadiusMin
    real(DP)                  :: RadiusMax
    real(DP)                  :: SizeParamMin
    real(DP)                  :: SizeParamMax

    real(DP)                  :: Radius
    real(DP)                  :: RadiusS
    real(DP)                  :: RadiusE
    real(DP)                  :: Weight
    real(DP), allocatable     :: a_GQRadius(:)
    real(DP), allocatable     :: a_GQWeight(:)

    integer                   :: NRadius
    integer                   :: NRadiusMax
    real(DP)    , allocatable :: ra_ParticleEffRadius(:,:)

    integer                   :: iWaveNum
    integer                   :: iRefIndex
    integer                   :: l
    integer                   :: m
    integer                   :: n

    character(stdstr) :: H2ODataDir = "testdir/data"



    NWaveNum   = int( ( WaveNumE - WaveNumS ) / DelWaveNum ) + 1
!!$    DelWaveNumMod = ( WaveNumE - WaveNumS ) / dble( NWaveNum )
    DelWaveNumMod = DelWaveNum

    allocate( w_WaveNum   ( NWaveNum ) )
    allocate( w_Wavelength( NWaveNum ) )

    allocate( w_RayScatCoefNonRadAct( NWaveNum ) )
    allocate( wm_RayScatCoef( NWaveNum, NMolNum ) )



    NRadiusMax = 0
    do n = 1, NParticle
      NRadius = a_ParticleNEffRadius(n)
      NRadiusMax = max( NRadiusMax, NRadius )
    end do
    allocate( ra_ParticleEffRadius( NRadiusMax, NParticle ) )
    do n = 1, NParticle
      NRadius = a_ParticleNEffRadius(n)
      RadiusS = a_ParticleEffRadiusS(n)
      RadiusE = a_ParticleEffRadiusE(n)
      do l = 1, NRadius
        if ( NRadius == 1 ) then
          Radius = RadiusS
        else
!!$          Radius = &
!!$            &   ( RadiusE - RadiusS ) &
!!$            & / ( NRadius - 1       ) &
!!$            & * ( l       - 1       ) &
!!$            & + RadiusS
          Radius = &
            &   log( RadiusE / RadiusS ) &
            & / ( NRadius - 1 ) &
            & * ( l       - 1 ) &
            & + log( RadiusS )
          Radius = exp( Radius )
        end if
        ra_ParticleEffRadius(l,n) = Radius
      end do
      do l = NRadius+1, NRadiusMax
        ra_ParticleEffRadius(l,n) = MissingValue
      end do
    end do

    allocate( wra_QExt  ( NWaveNum, NRadiusMax, NParticle ) )
    allocate( wra_SSA   ( NWaveNum, NRadiusMax, NParticle ) )
    allocate( wra_AF    ( NWaveNum, NRadiusMax, NParticle ) )


    call particle_ec_io_output_init( &
      & OutputFileName, NWaveNum, &
      & MolNumNonRadAct, &
      & NMolNum, m_MolNum, &
      & NParticle, a_ParticleName, a_ParticleNEffRadius, &
      & NRadiusMax, ra_ParticleEffRadius, &
      & MissingValue, DelWaveNum &
      & )


    !
    ! loop for wavenumber
    !

    do iWaveNum = 1, NWaveNum
      w_WaveNum( iWaveNum ) = &
        & WaveNumS + DelWaveNumMod * dble( iWaveNum - 1 )
    end do
    w_Wavelength = 1.0d0 / w_WaveNum

!!$    m = 0
    m = MolNumNonRadAct
    do iWaveNum = 1, NWaveNum
      call CalcRayScatCoef( &
        & m, w_Wavelength(iWaveNum), &
        & w_RayScatCoefNonRadAct(iWaveNum) &
        & )
    end do
    do m = 1, NMolNum
      do iWaveNum = 1, NWaveNum
        call CalcRayScatCoef( &
          & m_MolNum(m), w_Wavelength(iWaveNum), &
          & wm_RayScatCoef(iWaveNum,m) &
          & )
      end do
    end do

    wra_QExt = MissingValue
    wra_SSA  = MissingValue
    wra_AF   = MissingValue
    !
    do n = 1, NParticle

      if ( a_FlagReadQExtSSAAFFromFile(n) ) then

        if ( a_ParticleNEffRadius(n) /= 1 ) then
          write( 6, * ) 'ParticleNEffRadius has to be 1, when FlagReadOptPropFromFile is .true..'
          stop
        end if

        do l = 1, a_ParticleNEffRadius(n)
          call ParticleECCalcSetOptProp( &
            & a_ParticleDataFN(n), &
            & n, &
            & NWaveNum, w_Wavelength, &
            & wra_QExt(:,l,n), wra_SSA(:,l,n), wra_AF(:,l,n) &
            & )
        end do

      else if ( a_FlagReadQExtSSAAFFromParam(n) ) then

        select case ( a_ParticleName(n) )
        case ( 'water' )

          call H2OWaterInit( H2ODataDir )

          do l = 1, a_ParticleNEffRadius(n)
            EffRadius = ra_ParticleEffRadius(l,n)
            do iWaveNum = 1, NWaveNum
              call H2OWaterCalcParam( &
                & EffRadius, w_Wavelength(iWaveNum), &
                & wra_QExt(iWaveNum,l,n), &
                & wra_SSA (iWaveNum,l,n), &
                & wra_AF  (iWaveNum,l,n)  &
                & )
            end do
          end do

          call H2OWaterFinalize

        case ( 'ice' )

          call H2OIceInit  ( H2ODataDir )

          do l = 1, a_ParticleNEffRadius(n)
            EffRadius = ra_ParticleEffRadius(l,n)
            do iWaveNum = 1, NWaveNum
              call H2OIceCalcParam( &
                & EffRadius, w_Wavelength(iWaveNum), &
                & wra_QExt(iWaveNum,l,n), &
                & wra_SSA (iWaveNum,l,n), &
                & wra_AF  (iWaveNum,l,n)  &
                & )
!!$              if ( l == 50 ) then
!!$                write( 60, * ) w_Wavelength(iWaveNum), wra_SSA(iWaveNum,l,n)
!!$              end if
            end do
          end do

          call H2OIceFinalize

        case default

          write( 6, * ) 'Unexpected particle name for current version: ', &
            trim( a_ParticleName(n) )
          stop

        end select

      else

        select case ( a_ParticleSDF(n) )
        case ( 'marshall-palmer' )
          IDSizeDistFunc = IDSizeDistFuncMarshallPalmer
        case ( 'gamma'   )
          IDSizeDistFunc = IDSizeDistFuncGamma
!!$        case ( 'modified-gamma' )
!!$          IDSizeDistFunc = IDSizeDistFuncModifiedGamma
        case ( 'lognormal' )
          IDSizeDistFunc = IDSizeDistFuncLogNormal
        case default
          write( 6, * ) 'Unexpected particle name for size distribution: ', &
            trim( a_ParticleSDF(n) )
          stop
        end select

        NRadiusInteg = a_ParticleNRadiusInteg(n)


        allocate( a_GQRadius( NRadiusInteg ) )
        allocate( a_GQWeight( NRadiusInteg ) )


        call RefIndexInit( &
          & a_ParticleDataFN(n), &
          & FlagChkRangeAndStop &
          & )


        do iWaveNum = 1, NWaveNum
          if ( ( iWaveNum == 1 ) .or. ( mod( iWaveNum, NWaveNum/10 ) == 0 ) ) then
            write( 6, * ) '  Wavenumber loop : 1 <= ', iWaveNum, ' <= ', NWaveNum
          end if

          RadiusIntegWidth = a_ParticleRadiusIntegWidth(n)
          l = 1
          EffRadius  = ra_ParticleEffRadius(l,n)
          RadiusMin  = max( EffRadius - RadiusIntegWidth, 1.0d-10 )
          l = a_ParticleNEffRadius(n)
          EffRadius  = ra_ParticleEffRadius(l,n)
          RadiusMax  =      EffRadius + RadiusIntegWidth

          ! X=2.0D0*PI*rad*Rmed/wl
          SizeParamMin = 2.0d0 * PI * Rmed * RadiusMin / w_Wavelength(iWaveNum)
          SizeParamMax = 2.0d0 * PI * Rmed * RadiusMax / w_Wavelength(iWaveNum)
          NRefRadius = int( log10( SizeParamMax / SizeParamMin ) ) * 100 + 1
!!$          if ( ( iWaveNum == 1 ) .or. ( mod( iWaveNum, NWaveNum/10 ) == 0 ) ) then
          if ( iWaveNum == 1 ) then
            write( 6, * ) '    NRefRadius : ', NRefRadius
          end if
          allocate( r_RefRadius( NRefRadius ) )
          allocate( r_RefQExt  ( NRefRadius ) )
          allocate( r_RefSSA   ( NRefRadius ) )
          allocate( r_RefAF    ( NRefRadius ) )


          iRefIndex = 0
          do iRadius = 1, NRefRadius
            Radius = &
              &    log( RadiusMax  / RadiusMin ) &
              & / dble( NRefRadius - 1         ) &
              & * dble( iRadius    - 1         ) &
              & + log( RadiusMin )
            Radius = exp( Radius )

            call RefIndexCalc( &
              & w_Wavelength(iWaveNum),    & ! (in)
              & RIReal, RIImag,      & ! (out)
              & iRefIndex            &
              & )

            call CALLBH3(                &
              & Rmed, RIreal, RIimag, Radius, w_Wavelength(iWaveNum), &
              & QExt, SSA, SSCoA, AF &
              & )

            r_RefRadius(iRadius) = Radius
            r_RefQExt  (iRadius) = QExt
            r_RefSSA   (iRadius) = SSA
            r_RefAF    (iRadius) = AF
          end do


          do l = 1, a_ParticleNEffRadius(n)
            EffRadius = ra_ParticleEffRadius(l,n)

            RadiusIntegWidth = a_ParticleRadiusIntegWidth(n)

            RadiusS   = max( EffRadius - RadiusIntegWidth, 0.0d0 )
            RadiusE   =      EffRadius + RadiusIntegWidth

            if ( iWaveNum == 1 ) then
              write( 6, * ) '  Effective Radius count : 1 <= ', l, ' <= ', a_ParticleNEffRadius(n)
              if ( l == 1 ) then
                write( 6, * ) '  Particle name : ', trim( a_ParticleName(n) )
                write( 6, * ) '    Particle SDF : ', trim( a_ParticleSDF(n) )
              end if
!!$              write( 6, * ) '    Effective radius : ', EffRadius
            end if

            select case ( IDSizeDistFunc )
            case ( IDSizeDistFuncMarshallPalmer )
              Slope     = 3.0d0 / EffRadius

              if ( iWaveNum == 1 ) then
                write( 6, * ) '      Slope recip. : ', 1.0d0/Slope
                write( 6, * ) '      Radius range : ', RadiusS, RadiusE
                write( 6, * ) '      NRadiusInteg : ', NRadiusInteg
              end if
            case ( IDSizeDistFuncGamma )
              ! Heymsfield et al. (2013)
              Slope = ( 0.9d0 / ( 2.0d0**0.308d0 ) / (EffRadius*1.0d2) )**(1.0d0/0.692d0)
              Mu    = 0.9d0 / ( 2.0d0**0.308d0 ) * Slope**0.308 + 3.0d0
              Slope = Slope * 1.0d2

              if ( iWaveNum == 1 ) then
                write( 6, * ) '      Slope recip. : ', 1.0d0/Slope
                write( 6, * ) '      Dispersion   : ', Mu
                write( 6, * ) '      Radius range : ', RadiusS, RadiusE
                write( 6, * ) '      NRadiusInteg : ', NRadiusInteg
              end if
!!$          case ( IDSizeDistFuncModifiedGamma )
!!$            Mu    = a_SDFVariance    (n)
!!$            if ( Mu <= 0.0d0 ) then
!!$              write( 6, * ) 'Variance has to be positive: ', Mu, '.'
!!$              stop
!!$            end if
!!$            !
!!$            RadiusS   = a_RadiusIntegS   (n)
!!$            RadiusE   = a_RadiusIntegE   (n)
!!$
!!$            write( 6, * ) '      Effective radius : ', EffRadius
!!$            write( 6, * ) '      Variance         : ', Mu
!!$            write( 6, * ) '      Radius range     : ', RadiusS, RadiusE
!!$            write( 6, * ) '      NRadiusInteg     : ', NRadiusInteg
            case ( IDSizeDistFuncLogNormal )
              Mu    = a_ParticleStdDev(n)
              if ( Mu <= 0.0d0 ) then
                write( 6, * ) 'Standard deviation has to be positive: ', Mu, '.'
                stop
              end if
              !
              call CalcParamLogNormal( &
                & EffRadius, Mu, &
                & MedianRadius, ModeRadius, lnS  &
                & )

              if ( iWaveNum == 1 ) then
                write( 6, * ) '      Effective radius   : ', EffRadius
                write( 6, * ) '      Standard deviation : ', Mu
                write( 6, * ) '      Median radius      : ', MedianRadius
                write( 6, * ) '      Mode radius        : ', ModeRadius
                write( 6, * ) '      S                  : ', exp( lnS )
                write( 6, * ) '      Radius range       : ', RadiusS, RadiusE
                write( 6, * ) '      NRadiusInteg       : ', NRadiusInteg
              end if
            end select

            if ( ( EffRadius < RadiusS ) .or. ( RadiusE < EffRadius ) ) then
              write( 6, * ) 'Effective radius is out of radius range.'
              stop
            end if
            if ( ( MedianRadius < RadiusS ) .or. ( RadiusE < MedianRadius ) ) then
              write( 6, * ) 'Median radius is out of radius range.'
              stop
            end if
            if ( ( ModeRadius < RadiusS ) .or. ( RadiusE < ModeRadius ) ) then
              write( 6, * ) 'Mode radius is out of radius range.'
              stop
            end if


            if ( NRadiusInteg == 1 ) then
              a_GQRadius = EffRadius
              a_GQWeight = 1.0d0
            else
              call GauLeg( RadiusS, RadiusE, NRadiusInteg, a_GQRadius, a_GQWeight )
            end if

            wra_QExt(iWaveNum,l,n) = 0.0d0
            wra_SSA (iWaveNum,l,n) = 0.0d0
            wra_AF  (iWaveNum,l,n) = 0.0d0
            Sum1                   = 0.0d0
            Sum2                   = 0.0d0
            Sum3                   = 0.0d0
            do iRadius = 1, NRadiusInteg

              Radius = a_GQRadius(iRadius)
              Weight = a_GQWeight(iRadius)

              select case ( IDSizeDistFunc )
              case ( IDSizeDistFuncMarshallPalmer )
                ! Marshall-Palmer distribution
                SDF = SizeDistMarshallPalmer( Slope, Radius )
              case ( IDSizeDistFuncGamma )
                ! Gamma distribution
                SDF = SizeDistGamma( Mu, Slope, Radius )
!!$              case ( IDSizeDistFuncModifiedGamma )
!!$                ! modified Gamma distribution
!!$                SDF = SizeDistModifiedGamma( EffRadius, Mu, Radius )
              case ( IDSizeDistFuncLogNormal )
                ! lognormal distribution
                SDF = SizeDistLogNormal( MedianRadius, lnS, Radius )
              end select

              call InterpolateParams( &
                & NRefRadius, r_RefRadius, r_RefQExt, r_RefSSA, r_RefAF, &
                & Radius, &
                & QExt, SSA, AF &
                & )

              if ( QExt >= 0.0d0 ) then
                wra_QExt(iWaveNum,l,n) = wra_QExt(iWaveNum,l,n) &
                  & + QExt * Radius**2 * SDF * Weight
                Sum1                   = Sum1                   &
                  & +        Radius**2 * SDF * Weight
                wra_SSA (iWaveNum,l,n) = wra_SSA (iWaveNum,l,n) &
                  & + SSA  * QExt * Radius**2 * SDF * Weight
                Sum2                   = Sum2                   &
                  & +        QExt * Radius**2 * SDF * Weight
                wra_AF  (iWaveNum,l,n) = wra_AF  (iWaveNum,l,n) &
                  & + AF   * SSA  * QExt * Radius**2 * SDF * Weight
                Sum3                   = Sum3                   &
                  & +        SSA  * QExt * Radius**2 * SDF * Weight
              end if
            end do

            if ( Sum1 > 0.0d-100 ) then
              wra_QExt(iWaveNum,l,n) = wra_QExt(iWaveNum,l,n) / Sum1
            else
              wra_QExt(iWaveNum,l,n) = 0.0d0
            end if
            if ( Sum2 > 0.0d-100 ) then
              wra_SSA (iWaveNum,l,n) = wra_SSA (iWaveNum,l,n) / Sum2
            else
              wra_SSA (iWaveNum,l,n) = 0.0d0
            end if
            if ( Sum3 > 0.0d-100 ) then
              wra_AF  (iWaveNum,l,n) = wra_AF  (iWaveNum,l,n) / Sum3
            else
              wra_AF  (iWaveNum,l,n) = 0.0d0
            end if

          end do ! l EffRadius

          deallocate( r_RefRadius )
          deallocate( r_RefQExt   )
          deallocate( r_RefSSA    )
          deallocate( r_RefAF     )

        end do ! iWaveNum

        call RefIndexFinalize

        deallocate( a_GQRadius )
        deallocate( a_GQWeight )

      end if

    end do


    call particle_ec_io_output( &
      & NWaveNum, w_WaveNum, w_RayScatCoefNonRadAct, NMolNum, wm_RayScatCoef, &
      & NRadiusMax, NParticle, a_ParticleName, wra_QExt, wra_SSA, wra_AF &
      & )

    call particle_ec_io_output_end


    deallocate( ra_ParticleEffRadius )

    deallocate( wm_RayScatCoef )
    deallocate( wra_QExt  )
    deallocate( wra_SSA   )
    deallocate( wra_AF    )


  end subroutine ParticleECCalcSerial

  !----------------------------------------------------------------------------

  subroutine ParticleECCalcOMP( &
    & DelWaveNum, WaveNumS, WaveNumE, &
    & MolNumNonRadAct, &
    & NMolNum, m_MolNum, &
    & NParticle, &
    & a_FlagReadQExtSSAAFFromFile, &
    & a_FlagReadQExtSSAAFFromParam, &
    & a_ParticleName, a_ParticleSDF, a_ParticleDataFN, &
    & a_ParticleNEffRadius, a_ParticleEffRadiusS, a_ParticleEffRadiusE, &
    & a_ParticleStdDev, &
    & a_ParticleNRadiusInteg, &
    & OutputFileName &
    & )

    ! OpenMP
    !
    !$ use omp_lib

    use constants0, only: &
      & PI

    use rayscat, only : CalcRayScatCoef

    use refindex, only : &
      & RefIndexInit, &
      & RefIndexCalc, &
      & RefIndexFinalize

    use mie_theory, only : CALLBH3

    use particle_ec_io_module, only : &
      & particle_ec_io_output_init, &
      & particle_ec_io_output_end, &
      & particle_ec_io_output

    use h2o_water, only : &
      & H2OWaterInit, &
      & H2OWaterFinalize, &
      & H2OWaterCalcParam
    use h2o_ice, only : &
      & H2OIceInit, &
      & H2OIceFinalize, &
      & H2OIceCalcParam


    real(dp)    , intent(in ) :: DelWaveNum
    real(dp)    , intent(in ) :: WaveNumS
    real(dp)    , intent(in ) :: WaveNumE
    integer     , intent(in ) :: MolNumNonRadAct
    integer     , intent(in ) :: NMolNum
    integer     , intent(in ) :: m_MolNum(NMolNum)
    integer     , intent(in ) :: NParticle
    logical     , intent(in ) :: a_FlagReadQExtSSAAFFromFile (NParticle)
    logical     , intent(in ) :: a_FlagReadQExtSSAAFFromParam(NParticle)
    character(*), intent(in ) :: a_ParticleName            (NParticle)
    character(*), intent(in ) :: a_ParticleSDF             (NParticle)
    character(*), intent(in ) :: a_ParticleDataFN          (NParticle)
    integer     , intent(in ) :: a_ParticleNEffRadius      (NParticle)
    real(dp)    , intent(in ) :: a_ParticleEffRadiusS      (NParticle)
    real(dp)    , intent(in ) :: a_ParticleEffRadiusE      (NParticle)
    real(dp)    , intent(in ) :: a_ParticleStdDev          (NParticle)
    integer     , intent(in ) :: a_ParticleNRadiusInteg    (NParticle)
    character(*), intent(in ) :: OutputFileName


    ! a_ParticleName             : Particle name
    ! a_ParticleSDF              : Particle size distribution function
    ! a_ParticleDataFN           : Data filename for optical properties
    ! a_ParticleNEffRadius       : Number of effective radii
    ! a_ParticleEffRadiusS       : Start effective radius in a table
    ! a_ParticleEffRadiusE       : End   effective radius in a table
    ! a_ParticleStdDev           : Standard deviation of size distribution
    !                            : Valid only for lognormal distribution
    ! a_ParticleNRadiusInteg     : Number of points for radius integration


    ! Local variables
    !
    real(dp)    , parameter   :: MissingValue = -1.0d100

    integer(i4b)              :: NWaveNum
    real(dp)    , allocatable :: w_WaveNum   (:)
    real(dp)    , allocatable :: w_Wavelength(:)

    real(dp)                  :: DelWaveNumMod


    real(dp)    , allocatable :: w_RayScatCoefNonRadAct(:)
    real(dp)    , allocatable :: wm_RayScatCoef(:,:)
    real(dp)    , allocatable :: wra_QExt (:,:,:)
    real(dp)    , allocatable :: wra_SSA  (:,:,:)
    real(dp)    , allocatable :: wra_AF   (:,:,:)

    integer                   :: NRadiusInteg
    real(DP)                  :: EffRadius

    integer            :: IDSizeDistFunc

!!$    real(dp)    , parameter   :: Rmed = 1.00045D0
    real(dp)    , parameter   :: Rmed = 1.0d0

    real(DP)                  :: Radius
    real(DP)                  :: RadiusS
    real(DP)                  :: RadiusE

    integer                   :: NRadius
    integer                   :: NRadiusMax
    real(DP)    , allocatable :: ra_ParticleEffRadius(:,:)

    integer                   :: iWaveNum

    integer                   :: iThread
    integer                   :: NThreads
    integer     , allocatable :: a_iWaveNumS(:)
    integer     , allocatable :: a_iWaveNumE(:)

    integer                   :: l
    integer                   :: m
    integer                   :: n

    character(stdstr) :: H2ODataDir = "testdir/data"


    NWaveNum   = int( ( WaveNumE - WaveNumS ) / DelWaveNum ) + 1
!!$    DelWaveNumMod = ( WaveNumE - WaveNumS ) / dble( NWaveNum )
    DelWaveNumMod = DelWaveNum

    allocate( w_WaveNum   ( NWaveNum ) )
    allocate( w_Wavelength( NWaveNum ) )

    allocate( w_RayScatCoefNonRadAct( NWaveNum ) )
    allocate( wm_RayScatCoef( NWaveNum, NMolNum ) )


    NThreads = 1
    !$ NThreads  = omp_get_max_threads()
    !$ write( 6, * ) "Number of processors : ", omp_get_num_procs()
    !$ write( 6, * ) "Number of threads    : ", nthreads

    allocate( a_iWaveNumS( 0:NThreads-1 ) )
    allocate( a_iWaveNumE( 0:NThreads-1 ) )

    do iThread = 0, NThreads-1
      if ( iThread == 0 ) then
        a_iWaveNumS(iThread) = 1
      else
        a_iWaveNumS(iThread) = a_iWaveNumE(iThread-1) + 1
      end if
      a_iWaveNumE(iThread) = a_iWaveNumS(iThread) + NWaveNum / NThreads - 1
      if ( iThread + 1 <= mod( NWaveNum, NThreads ) ) then
        a_iWaveNumE(iThread) = a_iWaveNumE(iThread) + 1
      end if
    end do


    NRadiusMax = 0
    do n = 1, NParticle
      NRadius = a_ParticleNEffRadius(n)
      NRadiusMax = max( NRadiusMax, NRadius )
    end do
    allocate( ra_ParticleEffRadius( NRadiusMax, NParticle ) )
    do n = 1, NParticle
      NRadius = a_ParticleNEffRadius(n)
      RadiusS = a_ParticleEffRadiusS(n)
      RadiusE = a_ParticleEffRadiusE(n)
      do l = 1, NRadius
        if ( NRadius == 1 ) then
          Radius = RadiusS
        else
!!$          Radius = &
!!$            &   ( RadiusE - RadiusS ) &
!!$            & / ( NRadius - 1       ) &
!!$            & * ( l       - 1       ) &
!!$            & + RadiusS
          Radius = &
            &   log( RadiusE / RadiusS ) &
            & / ( NRadius - 1 ) &
            & * ( l       - 1 ) &
            & + log( RadiusS )
          Radius = exp( Radius )
        end if
        ra_ParticleEffRadius(l,n) = Radius
      end do
      do l = NRadius+1, NRadiusMax
        ra_ParticleEffRadius(l,n) = MissingValue
      end do
    end do

    allocate( wra_QExt  ( NWaveNum, NRadiusMax, NParticle ) )
    allocate( wra_SSA   ( NWaveNum, NRadiusMax, NParticle ) )
    allocate( wra_AF    ( NWaveNum, NRadiusMax, NParticle ) )


    call particle_ec_io_output_init( &
      & OutputFileName, NWaveNum, &
      & MolNumNonRadAct, &
      & NMolNum, m_MolNum, &
      & NParticle, a_ParticleName, a_ParticleNEffRadius, &
      & NRadiusMax, ra_ParticleEffRadius, &
      & MissingValue, DelWaveNum &
      & )


    !
    ! loop for wavenumber
    !

    do iWaveNum = 1, NWaveNum
      w_WaveNum( iWaveNum ) = &
        & WaveNumS + DelWaveNumMod * dble( iWaveNum - 1 )
    end do
    w_Wavelength = 1.0d0 / w_WaveNum

!!$    m = 0
    m = MolNumNonRadAct
    do iWaveNum = 1, NWaveNum
      call CalcRayScatCoef( &
        & m, w_Wavelength(iWaveNum), &
        & w_RayScatCoefNonRadAct(iWaveNum) &
        & )
    end do
    do m = 1, NMolNum
      do iWaveNum = 1, NWaveNum
        call CalcRayScatCoef( &
          & m_MolNum(m), w_Wavelength(iWaveNum), &
          & wm_RayScatCoef(iWaveNum,m) &
          & )
      end do
    end do

    wra_QExt = MissingValue
    wra_SSA  = MissingValue
    wra_AF   = MissingValue
    !
    do n = 1, NParticle

      if ( a_FlagReadQExtSSAAFFromFile(n) ) then

        if ( a_ParticleNEffRadius(n) /= 1 ) then
          write( 6, * ) 'ParticleNEffRadius has to be 1, when FlagReadOptPropFromFile is .true..'
          stop
        end if

        do l = 1, a_ParticleNEffRadius(n)
          call ParticleECCalcSetOptProp( &
            & a_ParticleDataFN(n), &
            & n, &
            & NWaveNum, w_Wavelength, &
            & wra_QExt(:,l,n), wra_SSA(:,l,n), wra_AF(:,l,n) &
            & )
        end do

      else if ( a_FlagReadQExtSSAAFFromParam(n) ) then

        select case ( a_ParticleName(n) )
        case ( 'water' )

          call H2OWaterInit( H2ODataDir )

          do l = 1, a_ParticleNEffRadius(n)
            EffRadius = ra_ParticleEffRadius(l,n)
            do iWaveNum = 1, NWaveNum
              call H2OWaterCalcParam( &
                & EffRadius, w_Wavelength(iWaveNum), &
                & wra_QExt(iWaveNum,l,n), &
                & wra_SSA (iWaveNum,l,n), &
                & wra_AF  (iWaveNum,l,n)  &
                & )
            end do
          end do

          call H2OWaterFinalize

        case ( 'ice' )

          call H2OIceInit  ( H2ODataDir )

          do l = 1, a_ParticleNEffRadius(n)
            EffRadius = ra_ParticleEffRadius(l,n)
            do iWaveNum = 1, NWaveNum
              call H2OIceCalcParam( &
                & EffRadius, w_Wavelength(iWaveNum), &
                & wra_QExt(iWaveNum,l,n), &
                & wra_SSA (iWaveNum,l,n), &
                & wra_AF  (iWaveNum,l,n)  &
                & )
!!$              if ( l == 50 ) then
!!$                write( 60, * ) w_Wavelength(iWaveNum), wra_SSA(iWaveNum,l,n)
!!$              end if
            end do
          end do

          call H2OIceFinalize

        case default

          write( 6, * ) 'Unexpected particle name for current version: ', &
            trim( a_ParticleName(n) )
          stop

        end select

      else

        select case ( a_ParticleSDF(n) )
        case ( 'marshall-palmer' )
          IDSizeDistFunc = IDSizeDistFuncMarshallPalmer
        case ( 'gamma'   )
          IDSizeDistFunc = IDSizeDistFuncGamma
!!$        case ( 'modified-gamma' )
!!$          IDSizeDistFunc = IDSizeDistFuncModifiedGamma
        case ( 'lognormal' )
          IDSizeDistFunc = IDSizeDistFuncLogNormal
        case default
          write( 6, * ) 'Unexpected particle name for size distribution: ', &
            trim( a_ParticleSDF(n) )
          stop
        end select

        NRadiusInteg = a_ParticleNRadiusInteg(n)



        call RefIndexInit( &
          & a_ParticleDataFN(n), &
          & FlagChkRangeAndStop &
          & )


!!$        do iWaveNum = 1, NWaveNum
        !$OMP PARALLEL DEFAULT(PRIVATE) &
        !$OMP SHARED( &
        !$OMP         NThreads, a_iWaveNumS, a_iWaveNumE, &
        !$OMP         NParticle,                                           &
        !$OMP         a_ParticleName, a_ParticleSDF, a_ParticleNEffRadius, &
        !$OMP         a_ParticleStdDev,                                    &
        !$OMP         NWaveNum, w_Wavelength,                              &
        !$OMP         NRadiusInteg, NRadiusMax, ra_ParticleEffRadius,      &
        !$OMP         IDSizeDistFunc, n,                                   &
        !$OMP         wra_QExt, wra_SSA, wra_AF                            &
        !$OMP       )

        !$OMP DO
        do iThread = 0, NThreads-1
          do iWaveNum = a_iWaveNumS(iThread), a_iWaveNumE(iThread)

            if ( iThread == 0 ) then
              if ( &
                &    ( iWaveNum == a_iWaveNumS(iThread) ) .or. &
                &    ( mod( iWaveNum, (a_iWaveNumE(iThread)-a_iWaveNumS(iThread))/10 ) == 0 ) &
                & ) then
                write( 6, * ) &
                  & '  Wavenumber loop : ', a_iWaveNumS(iThread), &
                  & ' <= ', iWaveNum, &
                  & ' <= ', a_iWaveNumE(iThread)
              end if
            end if

            call ParticleECCalcCore( &
              & NParticle,                                           & ! (in)
              & a_ParticleName, a_ParticleSDF, a_ParticleNEffRadius, & ! (in)
              & a_ParticleStdDev,                                    & ! (in)
              & NWaveNum, w_Wavelength,                              & ! (in)
              & NRadiusInteg, NRadiusMax, ra_ParticleEffRadius,      & ! (in)
              & IDSizeDistFunc, n,                                   & ! (in)
              & iWaveNum,                                            & ! (in)
              & wra_QExt, wra_SSA, wra_AF                            & ! (inout)
              & )

          end do ! iWaveNum

        end do
        !$OMP END DO
        !$OMP END PARALLEL

        call RefIndexFinalize


      end if

    end do


    call particle_ec_io_output( &
      & NWaveNum, w_WaveNum, w_RayScatCoefNonRadAct, NMolNum, wm_RayScatCoef, &
      & NRadiusMax, NParticle, a_ParticleName, wra_QExt, wra_SSA, wra_AF &
      & )

    call particle_ec_io_output_end


    deallocate( ra_ParticleEffRadius )

    deallocate( wm_RayScatCoef )
    deallocate( wra_QExt  )
    deallocate( wra_SSA   )
    deallocate( wra_AF    )

    deallocate( a_iWaveNumS )
    deallocate( a_iWaveNumE )


  end subroutine ParticleECCalcOMP

  !----------------------------------------------------------------------------

  subroutine ParticleECCalcCore( &
    & NParticle,                                           & ! (in)
    & a_ParticleName, a_ParticleSDF, a_ParticleNEffRadius, & ! (in)
    & a_ParticleStdDev,                                    & ! (in)
    & NWaveNum, w_Wavelength,                              & ! (in)
    & NRadiusInteg, NRadiusMax, ra_ParticleEffRadius,      & ! (in)
    & IDSizeDistFunc, iParticle,                           & ! (in)
    & iWaveNum,                                            & ! (in)
    & wra_QExt, wra_SSA, wra_AF                            & ! (inout)
    & )

    use constants0, only: &
      & PI

    use gauss_quad, only : GauLeg

    use refindex, only : RefIndexCalc

    use mie_theory, only : CALLBH3


    integer     , intent(in   ) :: NParticle
    character(*), intent(in   ) :: a_ParticleName            (NParticle)
    character(*), intent(in   ) :: a_ParticleSDF             (NParticle)
    integer     , intent(in   ) :: a_ParticleNEffRadius      (NParticle)
    real(dp)    , intent(in   ) :: a_ParticleStdDev          (NParticle)
    integer(i4b), intent(in   ) :: NWaveNum
    real(dp)    , intent(in   ) :: w_Wavelength        (NWaveNum)
    integer     , intent(in   ) :: NRadiusInteg
    integer     , intent(in   ) :: NRadiusMax
    real(DP)    , intent(in   ) :: ra_ParticleEffRadius(NRadiusMax, NParticle)
    integer     , intent(in   ) :: IDSizeDistFunc
    integer     , intent(in   ) :: iParticle
    integer     , intent(in   ) :: iWaveNum
    real(dp)    , intent(inout) :: wra_QExt(NWaveNum,NRadiusMax,NParticle)
    real(dp)    , intent(inout) :: wra_SSA (NWaveNum,NRadiusMax,NParticle)
    real(dp)    , intent(inout) :: wra_AF  (NWaveNum,NRadiusMax,NParticle)


    ! a_ParticleName             : Particle name
    ! a_ParticleSDF              : Particle size distribution function
    ! a_ParticleDataFN           : Data filename for optical properties
    ! a_ParticleNEffRadius       : Number of effective radii
    ! a_ParticleEffRadiusS       : Start effective radius in a table
    ! a_ParticleEffRadiusE       : End   effective radius in a table
    ! a_ParticleStdDev           : Standard deviation of size distribution
    !                            : Valid only for lognormal distribution
    ! a_ParticleNRadiusInteg     : Number of points for radius integration


    ! Local variables
    !
    integer                   :: NRadiusIntegLocal

    integer                   :: iRadius
    real(DP)                  :: EffRadius
    real(DP)                  :: QExt
    real(DP)                  :: SSA
    real(DP)                  :: SSCoA
    real(DP)                  :: AF

    real(DP)                  :: Radius2ndMoment
    real(DP)                  :: SDF


    ! variables for size distribution
!!$    real(DP)                  :: Slope
    ! variables for Gamma distribution
    real(DP)                  :: Mu
    !
    real(DP)                  :: MedianRadius
    real(DP)                  :: ModeRadius
    real(DP)                  :: lnS
    !
    real(DP)                  :: a_lnS         (a_ParticleNEffRadius(iParticle))
    real(DP)                  :: a_MedianRadius(a_ParticleNEffRadius(iParticle))
    !
!!$    real(dp)    , parameter   :: Rmed = 1.00045D0
    real(dp)    , parameter   :: Rmed = 1.0d0

    real(DP)                  :: RIReal
    real(DP)                  :: RIImag

    integer(i4b)              :: NRefRadius
    real(DP), allocatable     :: r_RefRadius(:)
    real(DP), allocatable     :: r_RefQExt  (:)
    real(DP), allocatable     :: r_RefSSA   (:)
    real(DP), allocatable     :: r_RefAF    (:)

    real(DP)                  :: RadiusMin
    real(DP)                  :: RadiusMax
    real(DP)                  :: SizeParamMin
    real(DP)                  :: SizeParamMax

    real(DP)                  :: a_GQRadius(NRadiusInteg)
    real(DP)                  :: a_GQWeight(NRadiusInteg)
    real(DP)                  :: Radius
    real(DP)                  :: RadiusS
    real(DP)                  :: RadiusE
    real(DP)                  :: lnRadiusS
    real(DP)                  :: lnRadiusE

    real(DP)                  :: a_RadiusS  (a_ParticleNEffRadius(iParticle))
    real(DP)                  :: a_RadiusE  (a_ParticleNEffRadius(iParticle))
    real(DP)                  :: a_lnRadiusS(a_ParticleNEffRadius(iParticle))
    real(DP)                  :: a_lnRadiusE(a_ParticleNEffRadius(iParticle))

    real(DP)                  :: Weight

    integer                   :: iRefIndex

    integer                   :: l
    integer                   :: n            ! index for do n = 1, NParticle


    n = iParticle

    !----------------------------------------------------
    ! Setup a table for QExt, SSA, and AF
    !   Making this table will save computational time.
    !
    select case ( IDSizeDistFunc )
    case ( IDSizeDistFuncLogNormal )

      RadiusMin  = 1.0d100
      RadiusMax  = 0.0d0

      Mu = a_ParticleStdDev(n)
      if ( Mu <= 0.0d0 ) then
        write( 6, * ) 'Standard deviation has to be positive: ', Mu, '.'
        stop
      end if
      !
      l = 1
      EffRadius   = ra_ParticleEffRadius(l,n)
      call CalcParamLogNormal( &
        & EffRadius, Mu, &
        & MedianRadius, ModeRadius, lnS  &
        & )
      call SetRadiusIntegRangeLogNormal(                 &
        & log( EffRadius ), log( MedianRadius ), lnS,    & ! (in)
        & lnRadiusS, lnRadiusE                           & ! (out)
        & )
      RadiusMin  = min( RadiusMin, exp( lnRadiusS ) )
      RadiusMax  = max( RadiusMax, exp( lnRadiusE ) )
      !
      l = a_ParticleNEffRadius(n)
      EffRadius   = ra_ParticleEffRadius(l,n)
      call CalcParamLogNormal( &
        & EffRadius, Mu, &
        & MedianRadius, ModeRadius, lnS  &
        & )
      call SetRadiusIntegRangeLogNormal(                 &
        & log( EffRadius ), log( MedianRadius ), lnS,    & ! (in)
        & lnRadiusS, lnRadiusE                           & ! (out)
        & )
      RadiusMin  = min( RadiusMin, exp( lnRadiusS ) )
      RadiusMax  = max( RadiusMax, exp( lnRadiusE ) )

    case default

      write( 6, * ) &
        & 'Unsupported size distribution function: ', IDSizeDistFunc
      stop


!!$      RadiusIntegWidth = a_ParticleRadiusIntegWidth(n)
!!$      l = 1
!!$      EffRadius  = ra_ParticleEffRadius(l,n)
!!$      RadiusMin  = max( EffRadius - RadiusIntegWidth, 1.0d-10 )
!!$      l = a_ParticleNEffRadius(n)
!!$      EffRadius  = ra_ParticleEffRadius(l,n)
!!$      RadiusMax  =      EffRadius + RadiusIntegWidth

    end select

    ! X=2.0D0*PI*rad*Rmed/wl
    SizeParamMin = 2.0d0 * PI * Rmed * RadiusMin / w_Wavelength(iWaveNum)
    SizeParamMax = 2.0d0 * PI * Rmed * RadiusMax / w_Wavelength(iWaveNum)
    NRefRadius = int( log10( SizeParamMax / SizeParamMin ) ) * 100 + 1
    if ( iWaveNum == 1 ) then
      write( 6, * ) '    NRefRadius : ', NRefRadius
    end if
    allocate( r_RefRadius( NRefRadius ) )
    allocate( r_RefQExt  ( NRefRadius ) )
    allocate( r_RefSSA   ( NRefRadius ) )
    allocate( r_RefAF    ( NRefRadius ) )

    iRefIndex = 0
    do iRadius = 1, NRefRadius
      if ( NRefRadius == 1 ) then
        Radius = RadiusMin
      else
        Radius = &
          &    log( RadiusMax  / RadiusMin ) &
          & / dble( NRefRadius - 1         ) &
          & * dble( iRadius    - 1         ) &
          & + log( RadiusMin )
        Radius = exp( Radius )
      end if

      call RefIndexCalc( &
        & w_Wavelength(iWaveNum),    & ! (in)
        & RIReal, RIImag,      & ! (out)
        & iRefIndex            &
        & )

      call CALLBH3(                &
        & Rmed, RIreal, RIimag, Radius, w_Wavelength(iWaveNum), &
        & QExt, SSA, SSCoA, AF &
        & )

      r_RefRadius(iRadius) = Radius
      r_RefQExt  (iRadius) = QExt
      r_RefSSA   (iRadius) = SSA
      r_RefAF    (iRadius) = AF
    end do
    !----------------------------------------------------


    do l = 1, a_ParticleNEffRadius(n)

      NRadiusIntegLocal = NRadiusInteg

      if ( iWaveNum == 1 ) then
        write( 6, * ) '  Effective Radius count : 1 <= ', l, ' <= ', a_ParticleNEffRadius(n)
        if ( l == 1 ) then
          write( 6, * ) '  Particle name : ', trim( a_ParticleName(n) )
          write( 6, * ) '    Particle SDF : ', trim( a_ParticleSDF(n) )
        end if
      end if


      select case ( IDSizeDistFunc )
      case ( IDSizeDistFuncLogNormal )

        EffRadius   = ra_ParticleEffRadius(l,n)

        Mu = a_ParticleStdDev(n)
        if ( Mu <= 0.0d0 ) then
          write( 6, * ) 'Standard deviation has to be positive: ', Mu, '.'
          stop
        end if
        !
        call CalcParamLogNormal( &
          & EffRadius, Mu, &
          & MedianRadius, ModeRadius, lnS  &
          & )
        call SetRadiusIntegRangeLogNormal(                 &
          & log( EffRadius ), log( MedianRadius ), lnS,    & ! (in)
          & lnRadiusS, lnRadiusE                           & ! (out)
          & )
        if ( lnS == 0 ) then
          NRadiusIntegLocal = 1
        else
          NRadiusIntegLocal = NRadiusInteg
        end if
        RadiusS          = exp( lnRadiusS )
        RadiusE          = exp( lnRadiusE )


        if ( iWaveNum == 1 ) then
          write( 6, * ) '      Effective radius       : ', EffRadius
          write( 6, * ) '      ln( Effective radius ) : ', log( EffRadius )
          write( 6, * ) '      Standard deviation     : ', Mu
          write( 6, * ) '      Median radius          : ', MedianRadius
          write( 6, * ) '      Mode radius            : ', ModeRadius
          write( 6, * ) '      S                      : ', exp( lnS )
          write( 6, * ) '      lnS                    : ', lnS
          write( 6, * ) '      Radius range           : ', RadiusS, RadiusE
          write( 6, * ) '      ln(Radius) range       : ', lnRadiusS, lnRadiusE
          write( 6, * ) '      NRadiusInteg           : ', NRadiusIntegLocal
        end if


        if ( NRadiusIntegLocal > 1 ) then
          if ( ( EffRadius < RadiusS ) .or. ( RadiusE < EffRadius ) ) then
            write( 6, * ) 'ERROR: Effective radius is out of radius range.'
            write( 6, * ) '  EffRadius          : ', EffRadius
            write( 6, * ) '  (RadiusS, RadiusE) : (', RadiusS, ',', RadiusE, ')'
            stop
          end if
        end if
        if ( lnS > 0.0d0 ) then
          if ( ( MedianRadius < RadiusS ) .or. ( RadiusE < MedianRadius ) ) then
            write( 6, * ) 'WARNING: Median radius is out of radius range.'
            write( 6, * ) '  MedianRadius       : ', MedianRadius
            write( 6, * ) '  (RadiusS, RadiusE) : (', RadiusS, ',', RadiusE, ')'
            if ( NRadiusIntegLocal > 1 ) then
              stop
            end if
          end if
          if ( ( ModeRadius < RadiusS ) .or. ( RadiusE < ModeRadius ) ) then
            write( 6, * ) 'WARNING: Mode radius is out of radius range.'
            write( 6, * ) '  ModeRadius         : ', ModeRadius
            write( 6, * ) '  (RadiusS, RadiusE) : (', RadiusS, ',', RadiusE, ')'
            if ( NRadiusIntegLocal > 1 ) then
              stop
            end if
          end if
        end if


        if ( NRadiusIntegLocal == 1 ) then
          a_GQRadius(1)              = lnRadiusS
          a_GQWeight(1)              =  1.0d0
          a_GQRadius(2:NRadiusInteg) = -1.0d100
          a_GQWeight(2:NRadiusInteg) =  0.0d0
        else
          call GauLeg( lnRadiusS, lnRadiusE, &
            & NRadiusIntegLocal, a_GQRadius, a_GQWeight )
        end if

      case default

        write( 6, * ) &
          & 'Unsupported size distribution function: ', IDSizeDistFunc
        stop


        EffRadius = ra_ParticleEffRadius(l,n)

!!$        select case ( IDSizeDistFunc )
!!$        case ( IDSizeDistFuncMarshallPalmer )
!!$          Slope     = 3.0d0 / EffRadius
!!$
!!$          if ( iWaveNum == 1 ) then
!!$            write( 6, * ) '      Slope recip. : ', 1.0d0/Slope
!!$            write( 6, * ) '      Radius range : ', RadiusS, RadiusE
!!$            write( 6, * ) '      NRadiusInteg : ', NRadiusInteg
!!$          end if
!!$        case ( IDSizeDistFuncGamma )
!!$          ! Heymsfield et al. (2013)
!!$          Slope = ( 0.9d0 / ( 2.0d0**0.308d0 ) / (EffRadius*1.0d2) )**(1.0d0/0.692d0)
!!$          Mu    = 0.9d0 / ( 2.0d0**0.308d0 ) * Slope**0.308 + 3.0d0
!!$          Slope = Slope * 1.0d2
!!$
!!$          if ( iWaveNum == 1 ) then
!!$            write( 6, * ) '      Slope recip. : ', 1.0d0/Slope
!!$            write( 6, * ) '      Dispersion   : ', Mu
!!$            write( 6, * ) '      Radius range : ', RadiusS, RadiusE
!!$            write( 6, * ) '      NRadiusInteg : ', NRadiusInteg
!!$          end if
!!$        case ( IDSizeDistFuncModifiedGamma )
!!$          Mu    = a_SDFVariance    (n)
!!$          if ( Mu <= 0.0d0 ) then
!!$            write( 6, * ) 'Variance has to be positive: ', Mu, '.'
!!$            stop
!!$          end if
!!$          !
!!$          RadiusS   = a_RadiusIntegS   (n)
!!$          RadiusE   = a_RadiusIntegE   (n)
!!$
!!$          write( 6, * ) '      Effective radius : ', EffRadius
!!$          write( 6, * ) '      Variance         : ', Mu
!!$          write( 6, * ) '      Radius range     : ', RadiusS, RadiusE
!!$          write( 6, * ) '      NRadiusInteg     : ', NRadiusInteg
!!$        end select


!!$        if ( ( EffRadius < RadiusS ) .or. ( RadiusE < EffRadius ) ) then
!!$          write( 6, * ) 'Effective radius is out of radius range.'
!!$          stop
!!$        end if
!!$        if ( ( MedianRadius < RadiusS ) .or. ( RadiusE < MedianRadius ) ) then
!!$          write( 6, * ) 'Median radius is out of radius range.'
!!$          stop
!!$        end if
!!$        if ( ( ModeRadius < RadiusS ) .or. ( RadiusE < ModeRadius ) ) then
!!$          write( 6, * ) 'Mode radius is out of radius range.'
!!$          stop
!!$        end if


!!$        if ( NRadiusIntegLocal == 1 ) then
!!$          a_GQRadius(1)              = EffRadius
!!$          a_GQWeight(1)              = 1.0d0
!!$          a_GQRadius(2:NRadiusInteg) = 1.0d100
!!$          a_GQWeight(2:NRadiusInteg) = 0.0d0
!!$        else
!!$          call GauLeg( RadiusS, RadiusE, NRadiusIntegLocal, a_GQRadius, a_GQWeight )
!!$        end if

      end select



      Radius2ndMoment        = 0.0d0
      wra_QExt(iWaveNum,l,n) = 0.0d0
      wra_SSA (iWaveNum,l,n) = 0.0d0
      wra_AF  (iWaveNum,l,n) = 0.0d0
      do iRadius = 1, NRadiusIntegLocal


        select case ( IDSizeDistFunc )
!!$        case ( IDSizeDistFuncMarshallPalmer )
!!$          ! Marshall-Palmer distribution
!!$
!!$          Radius = a_GQRadius(iRadius)
!!$          Weight = a_GQWeight(iRadius)
!!$          SDF = SizeDistMarshallPalmer( Slope, Radius )
!!$
!!$        case ( IDSizeDistFuncGamma )
!!$          ! Gamma distribution
!!$
!!$          Radius = a_GQRadius(iRadius)
!!$          Weight = a_GQWeight(iRadius)
!!$          SDF = SizeDistGamma( Mu, Slope, Radius )
!!$        case ( IDSizeDistFuncModifiedGamma )
!!$          ! modified Gamma distribution
!!$
!!$          Radius = a_GQRadius(iRadius)
!!$          Weight = a_GQWeight(iRadius)
!!$          SDF = SizeDistModifiedGamma( EffRadius, Mu, Radius )

        case ( IDSizeDistFuncLogNormal )

          ! NOTE: a_GQRadius includes ln( radius ).
          Radius = exp( a_GQRadius(iRadius) )
          Weight = a_GQWeight(iRadius)

          ! lognormal distribution
          if ( lnS == 0.0d0 ) then
            SDF = 1.0d0
          else
            SDF = SizeDistLogNormal( MedianRadius, lnS, Radius )
            SDF = SDF * Radius
          end if

        end select

        call InterpolateParams( &
          & NRefRadius, r_RefRadius, r_RefQExt, r_RefSSA, r_RefAF, &
          & Radius, &
          & QExt, SSA, AF &
          & )

        if ( QExt >= 0.0d0 ) then
          Radius2ndMoment        = Radius2ndMoment        &
            & +        Radius**2 * SDF * Weight
          wra_QExt(iWaveNum,l,n) = wra_QExt(iWaveNum,l,n) &
            & + QExt * Radius**2 * SDF * Weight
          wra_SSA (iWaveNum,l,n) = wra_SSA (iWaveNum,l,n) &
            & + SSA  * QExt * Radius**2 * SDF * Weight
          wra_AF  (iWaveNum,l,n) = wra_AF  (iWaveNum,l,n) &
            & + AF   * SSA  * QExt * Radius**2 * SDF * Weight
        end if
      end do

      if ( wra_SSA (iWaveNum,l,n) > 0.0d-100 ) then
        wra_AF  (iWaveNum,l,n) = wra_AF  (iWaveNum,l,n) / wra_SSA (iWaveNum,l,n)
      else
        wra_AF  (iWaveNum,l,n) = 0.0d0
      end if
      if ( wra_QExt(iWaveNum,l,n) > 0.0d-100 ) then
        wra_SSA (iWaveNum,l,n) = wra_SSA (iWaveNum,l,n) / wra_QExt(iWaveNum,l,n)
      else
        wra_SSA (iWaveNum,l,n) = 0.0d0
      end if
      if ( Radius2ndMoment > 0.0d-100 ) then
        wra_QExt(iWaveNum,l,n) = wra_QExt(iWaveNum,l,n) / Radius2ndMoment
      else
        wra_QExt(iWaveNum,l,n) = 0.0d0
      end if

    end do ! l EffRadius

    deallocate( r_RefRadius )
    deallocate( r_RefQExt   )
    deallocate( r_RefSSA    )
    deallocate( r_RefAF     )


  end subroutine ParticleECCalcCore

  !----------------------------------------------------------------------------

  subroutine OLD01_ParticleECCalcCore( &
    & NParticle,                                           & ! (in)
    & a_ParticleName, a_ParticleSDF, a_ParticleNEffRadius, & ! (in)
    & a_ParticleStdDev, a_ParticleRadiusIntegWidth,        & ! (in)
    & NWaveNum, w_Wavelength,                              & ! (in)
    & NRadiusInteg, NRadiusMax, ra_ParticleEffRadius,      & ! (in)
    & IDSizeDistFunc, n,                                   & ! (in)
    & iWaveNum,                                            & ! (in)
    & wra_QExt, wra_SSA, wra_AF                            & ! (inout)
    & )

    use constants0, only: &
      & PI

    use gauss_quad, only : GauLeg

    use refindex, only : RefIndexCalc

    use mie_theory, only : CALLBH3


    integer     , intent(in   ) :: NParticle
    character(*), intent(in   ) :: a_ParticleName            (NParticle)
    character(*), intent(in   ) :: a_ParticleSDF             (NParticle)
    integer     , intent(in   ) :: a_ParticleNEffRadius      (NParticle)
    real(dp)    , intent(in   ) :: a_ParticleStdDev          (NParticle)
    real(dp)    , intent(in   ) :: a_ParticleRadiusIntegWidth(NParticle)
    integer(i4b), intent(in   ) :: NWaveNum
    real(dp)    , intent(in   ) :: w_Wavelength        (NWaveNum)
    integer     , intent(in   ) :: NRadiusInteg
    integer     , intent(in   ) :: NRadiusMax
    real(DP)    , intent(in   ) :: ra_ParticleEffRadius(NRadiusMax, NParticle)
    integer     , intent(in   ) :: IDSizeDistFunc
    integer     , intent(in   ) :: n            ! index for do n = 1, NParticle
    integer     , intent(in   ) :: iWaveNum
    real(dp)    , intent(inout) :: wra_QExt(NWaveNum,NRadiusMax,NParticle)
    real(dp)    , intent(inout) :: wra_SSA (NWaveNum,NRadiusMax,NParticle)
    real(dp)    , intent(inout) :: wra_AF  (NWaveNum,NRadiusMax,NParticle)


    ! a_ParticleName             : Particle name
    ! a_ParticleSDF              : Particle size distribution function
    ! a_ParticleDataFN           : Data filename for optical properties
    ! a_ParticleNEffRadius       : Number of effective radii
    ! a_ParticleEffRadiusS       : Start effective radius in a table
    ! a_ParticleEffRadiusE       : End   effective radius in a table
    ! a_ParticleStdDev           : Standard deviation of size distribution
    !                            : Valid only for lognormal distribution
    ! a_ParticleNRadiusInteg     : Number of points for radius integration
    ! a_ParticleRadiusIntegWidth : Radius width for radius integration


    ! Local variables
    !
    integer                   :: iRadius
    real(DP)                  :: EffRadius
    real(DP)                  :: RadiusIntegWidth
    real(DP)                  :: QExt
    real(DP)                  :: SSA
    real(DP)                  :: SSCoA
    real(DP)                  :: AF

    real(DP)                  :: Sum1
    real(DP)                  :: Sum2
    real(DP)                  :: Sum3
    real(DP)                  :: SDF


    ! variables for size distribution
    real(DP)                  :: Slope
    ! variables for Gamma distribution
    real(DP)                  :: Mu
    !
    real(DP)                  :: MedianRadius
    real(DP)                  :: ModeRadius
    real(DP)                  :: lnS
    !
!!$    real(dp)    , parameter   :: Rmed = 1.00045D0
    real(dp)    , parameter   :: Rmed = 1.0d0

    real(DP)                  :: RIReal
    real(DP)                  :: RIImag

    integer(i4b)              :: NRefRadius
    real(DP), allocatable     :: r_RefRadius(:)
    real(DP), allocatable     :: r_RefQExt  (:)
    real(DP), allocatable     :: r_RefSSA   (:)
    real(DP), allocatable     :: r_RefAF    (:)

    real(DP)                  :: RadiusMin
    real(DP)                  :: RadiusMax
    real(DP)                  :: SizeParamMin
    real(DP)                  :: SizeParamMax

    real(DP)                  :: a_GQRadius(NRadiusInteg)
    real(DP)                  :: a_GQWeight(NRadiusInteg)
    real(DP)                  :: Radius
    real(DP)                  :: RadiusS
    real(DP)                  :: RadiusE
    real(DP)                  :: Weight

    integer                   :: iRefIndex

    integer                   :: l


    RadiusIntegWidth = a_ParticleRadiusIntegWidth(n)
    l = 1
    EffRadius  = ra_ParticleEffRadius(l,n)
    RadiusMin  = max( EffRadius - RadiusIntegWidth, 1.0d-10 )
    l = a_ParticleNEffRadius(n)
    EffRadius  = ra_ParticleEffRadius(l,n)
    RadiusMax  =      EffRadius + RadiusIntegWidth

    ! X=2.0D0*PI*rad*Rmed/wl
    SizeParamMin = 2.0d0 * PI * Rmed * RadiusMin / w_Wavelength(iWaveNum)
    SizeParamMax = 2.0d0 * PI * Rmed * RadiusMax / w_Wavelength(iWaveNum)
    NRefRadius = int( log10( SizeParamMax / SizeParamMin ) ) * 100 + 1
!!$          if ( ( iWaveNum == 1 ) .or. ( mod( iWaveNum, NWaveNum/10 ) == 0 ) ) then
    if ( iWaveNum == 1 ) then
      write( 6, * ) '    NRefRadius : ', NRefRadius
    end if
    allocate( r_RefRadius( NRefRadius ) )
    allocate( r_RefQExt  ( NRefRadius ) )
    allocate( r_RefSSA   ( NRefRadius ) )
    allocate( r_RefAF    ( NRefRadius ) )


    iRefIndex = 0
    do iRadius = 1, NRefRadius
      Radius = &
        &    log( RadiusMax  / RadiusMin ) &
        & / dble( NRefRadius - 1         ) &
        & * dble( iRadius    - 1         ) &
        & + log( RadiusMin )
      Radius = exp( Radius )

      call RefIndexCalc( &
        & w_Wavelength(iWaveNum),    & ! (in)
        & RIReal, RIImag,      & ! (out)
        & iRefIndex            &
        & )

      call CALLBH3(                &
        & Rmed, RIreal, RIimag, Radius, w_Wavelength(iWaveNum), &
        & QExt, SSA, SSCoA, AF &
        & )

      r_RefRadius(iRadius) = Radius
      r_RefQExt  (iRadius) = QExt
      r_RefSSA   (iRadius) = SSA
      r_RefAF    (iRadius) = AF
    end do


    do l = 1, a_ParticleNEffRadius(n)
      EffRadius = ra_ParticleEffRadius(l,n)

      RadiusIntegWidth = a_ParticleRadiusIntegWidth(n)

      RadiusS   = max( EffRadius - RadiusIntegWidth, 0.0d0 )
      RadiusE   =      EffRadius + RadiusIntegWidth

      if ( iWaveNum == 1 ) then
        write( 6, * ) '  Effective Radius count : 1 <= ', l, ' <= ', a_ParticleNEffRadius(n)
        if ( l == 1 ) then
          write( 6, * ) '  Particle name : ', trim( a_ParticleName(n) )
          write( 6, * ) '    Particle SDF : ', trim( a_ParticleSDF(n) )
        end if
!!$              write( 6, * ) '    Effective radius : ', EffRadius
      end if

      select case ( IDSizeDistFunc )
      case ( IDSizeDistFuncMarshallPalmer )
        Slope     = 3.0d0 / EffRadius

        if ( iWaveNum == 1 ) then
          write( 6, * ) '      Slope recip. : ', 1.0d0/Slope
          write( 6, * ) '      Radius range : ', RadiusS, RadiusE
          write( 6, * ) '      NRadiusInteg : ', NRadiusInteg
        end if
      case ( IDSizeDistFuncGamma )
        ! Heymsfield et al. (2013)
        Slope = ( 0.9d0 / ( 2.0d0**0.308d0 ) / (EffRadius*1.0d2) )**(1.0d0/0.692d0)
        Mu    = 0.9d0 / ( 2.0d0**0.308d0 ) * Slope**0.308 + 3.0d0
        Slope = Slope * 1.0d2

        if ( iWaveNum == 1 ) then
          write( 6, * ) '      Slope recip. : ', 1.0d0/Slope
          write( 6, * ) '      Dispersion   : ', Mu
          write( 6, * ) '      Radius range : ', RadiusS, RadiusE
          write( 6, * ) '      NRadiusInteg : ', NRadiusInteg
        end if
!!$          case ( IDSizeDistFuncModifiedGamma )
!!$            Mu    = a_SDFVariance    (n)
!!$            if ( Mu <= 0.0d0 ) then
!!$              write( 6, * ) 'Variance has to be positive: ', Mu, '.'
!!$              stop
!!$            end if
!!$            !
!!$            RadiusS   = a_RadiusIntegS   (n)
!!$            RadiusE   = a_RadiusIntegE   (n)
!!$
!!$            write( 6, * ) '      Effective radius : ', EffRadius
!!$            write( 6, * ) '      Variance         : ', Mu
!!$            write( 6, * ) '      Radius range     : ', RadiusS, RadiusE
!!$            write( 6, * ) '      NRadiusInteg     : ', NRadiusInteg
      case ( IDSizeDistFuncLogNormal )
        Mu    = a_ParticleStdDev(n)
        if ( Mu <= 0.0d0 ) then
          write( 6, * ) 'Standard deviation has to be positive: ', Mu, '.'
          stop
        end if
        !
        call CalcParamLogNormal( &
          & EffRadius, Mu, &
          & MedianRadius, ModeRadius, lnS  &
          & )

        if ( iWaveNum == 1 ) then
          write( 6, * ) '      Effective radius   : ', EffRadius
          write( 6, * ) '      Standard deviation : ', Mu
          write( 6, * ) '      Median radius      : ', MedianRadius
          write( 6, * ) '      Mode radius        : ', ModeRadius
          write( 6, * ) '      S                  : ', exp( lnS )
          write( 6, * ) '      Radius range       : ', RadiusS, RadiusE
          write( 6, * ) '      NRadiusInteg       : ', NRadiusInteg
        end if
      end select

      if ( ( EffRadius < RadiusS ) .or. ( RadiusE < EffRadius ) ) then
        write( 6, * ) 'Effective radius is out of radius range.'
        stop
      end if
      if ( ( MedianRadius < RadiusS ) .or. ( RadiusE < MedianRadius ) ) then
        write( 6, * ) 'Median radius is out of radius range.'
        stop
      end if
      if ( ( ModeRadius < RadiusS ) .or. ( RadiusE < ModeRadius ) ) then
        write( 6, * ) 'Mode radius is out of radius range.'
        stop
      end if


      if ( NRadiusInteg == 1 ) then
        a_GQRadius = EffRadius
        a_GQWeight = 1.0d0
      else
        call GauLeg( RadiusS, RadiusE, NRadiusInteg, a_GQRadius, a_GQWeight )
      end if

      wra_QExt(iWaveNum,l,n) = 0.0d0
      wra_SSA (iWaveNum,l,n) = 0.0d0
      wra_AF  (iWaveNum,l,n) = 0.0d0
      Sum1                   = 0.0d0
      Sum2                   = 0.0d0
      Sum3                   = 0.0d0
      do iRadius = 1, NRadiusInteg

        Radius = a_GQRadius(iRadius)
        Weight = a_GQWeight(iRadius)

        select case ( IDSizeDistFunc )
        case ( IDSizeDistFuncMarshallPalmer )
          ! Marshall-Palmer distribution
          SDF = SizeDistMarshallPalmer( Slope, Radius )
        case ( IDSizeDistFuncGamma )
          ! Gamma distribution
          SDF = SizeDistGamma( Mu, Slope, Radius )
!!$              case ( IDSizeDistFuncModifiedGamma )
!!$                ! modified Gamma distribution
!!$                SDF = SizeDistModifiedGamma( EffRadius, Mu, Radius )
        case ( IDSizeDistFuncLogNormal )
          ! lognormal distribution
          SDF = SizeDistLogNormal( MedianRadius, lnS, Radius )
        end select

        call InterpolateParams( &
          & NRefRadius, r_RefRadius, r_RefQExt, r_RefSSA, r_RefAF, &
          & Radius, &
          & QExt, SSA, AF &
          & )

        if ( QExt >= 0.0d0 ) then
          wra_QExt(iWaveNum,l,n) = wra_QExt(iWaveNum,l,n) &
            & + QExt * Radius**2 * SDF * Weight
          Sum1                   = Sum1                   &
            & +        Radius**2 * SDF * Weight
          wra_SSA (iWaveNum,l,n) = wra_SSA (iWaveNum,l,n) &
            & + SSA  * QExt * Radius**2 * SDF * Weight
          Sum2                   = Sum2                   &
            & +        QExt * Radius**2 * SDF * Weight
          wra_AF  (iWaveNum,l,n) = wra_AF  (iWaveNum,l,n) &
            & + AF   * SSA  * QExt * Radius**2 * SDF * Weight
          Sum3                   = Sum3                   &
            & +        SSA  * QExt * Radius**2 * SDF * Weight
        end if
      end do

      if ( Sum1 > 0.0d-100 ) then
        wra_QExt(iWaveNum,l,n) = wra_QExt(iWaveNum,l,n) / Sum1
      else
        wra_QExt(iWaveNum,l,n) = 0.0d0
      end if
      if ( Sum2 > 0.0d-100 ) then
        wra_SSA (iWaveNum,l,n) = wra_SSA (iWaveNum,l,n) / Sum2
      else
        wra_SSA (iWaveNum,l,n) = 0.0d0
      end if
      if ( Sum3 > 0.0d-100 ) then
        wra_AF  (iWaveNum,l,n) = wra_AF  (iWaveNum,l,n) / Sum3
      else
        wra_AF  (iWaveNum,l,n) = 0.0d0
      end if

    end do ! l EffRadius

    deallocate( r_RefRadius )
    deallocate( r_RefQExt   )
    deallocate( r_RefSSA    )
    deallocate( r_RefAF     )


  end subroutine OLD01_ParticleECCalcCore

  !----------------------------------------------------------------------------

  subroutine ParticleECCalcSetOptProp( &
    & InNcFN, &
    & iParticle, &
    & NWavelen, w_Wavelen, &
    & w_QExt, w_SSA, w_AF &
    & )

    use ni3_module

    character(*), intent(in ) :: InNcFN
    integer     , intent(in ) :: iParticle
    integer     , intent(in ) :: NWavelen
    real(dp)    , intent(in ) :: w_Wavelen( NWavelen )
    real(dp)    , intent(out) :: w_QExt   ( NWavelen )
    real(dp)    , intent(out) :: w_SSA    ( NWavelen )
    real(dp)    , intent(out) :: w_AF     ( NWavelen )


    ! Local variables
    !
    character(stdstr) :: Mode
    character(stdstr) :: Name
    integer           :: NcID

    logical               :: FlagBnds

    integer               :: NPtcl
    integer               :: NData
    logical , allocatable :: w_FlagSet       (:)
    real(dp), allocatable :: aw_InWavelenBnds(:,:)
    real(dp), allocatable :: w_InWavelen(:)
    real(dp), allocatable :: wa_InQExt  (:,:)
    real(dp), allocatable :: wa_InSSA   (:,:)
    real(dp), allocatable :: wa_InAF    (:,:)

    real(dp), allocatable :: w_Tmp       (:)

    integer :: l
    integer :: m
    integer :: n


    allocate( w_FlagSet  ( NWavelen ) )

    Mode = "read"
    call ni3_open( InNcFN, Mode, NcID )

    Name = "wl"
    call ni3_inq_dimlen( NcID, Name, NData )

    Name = "PtclNum"
    call ni3_inq_dimlen( NcID, Name, NPtcl )

    ! check
    if ( ( iParticle < 1 ) .or. ( NPtcl < iParticle ) ) then
      write( 6, * ) 'Number of particles in ', trim( InNcFN ), ' is inappropriate: ', iParticle, NPtcl
      stop
    end if


    ! Now, FlagBnds is .false.
    FlagBnds = .false.
    !
!!$    Name = "wl_bnds"
!!$    FlagBnds = ni3_chk_var( NcID, Name )


    if ( FlagBnds ) then
       write( 6, * ) "A variable for wavenumber bounds, wl_bnds, is found."
       write( 6, * ) "Those are used to set optical parameters."
       allocate( aw_InWaveLenBnds( 2, NData ) )
       Name = "wl_bnds"
       call ni3_get_var( NcID, Name, aw_InWaveLenBnds )
       ! check
       do m = 1+1, NData
          if ( aw_InWaveLenBnds(2,m-1) /= aw_InWaveLenBnds(1,m) ) then
             write( 6, * ) "ERROR : Bounds does not match:"
             write( 6, * ) "ERROR : ", aw_InWaveLenBnds(1,m-1), aw_InWaveLenBnds(2,m-1)
             write( 6, * ) "ERROR : ", aw_InWaveLenBnds(1,m  ), aw_InWaveLenBnds(2,m  )
             stop
          end if
       end do
    else
       write( 6, * ) "A variable for wavenumber bounds, wl_bnds, is not found."
       write( 6, * ) "Wavenumber, wl, is used to set optical parameters."
       allocate( w_InWaveLen( NData ) )
       Name = "wl"
       call ni3_get_var( NcID, Name, w_InWaveLen )
    end if

    allocate( wa_InQExt  ( NData, NPtcl ) )
    allocate( wa_InSSA   ( NData, NPtcl ) )
    allocate( wa_InAF    ( NData, NPtcl ) )

    Name = "QExt"
    call ni3_get_var( NcID, Name, wa_InQExt   )
    Name = "SSA"
    call ni3_get_var( NcID, Name, wa_InSSA    )
    Name = "AF"
    call ni3_get_var( NcID, Name, wa_InAF     )

    call ni3_close( NcID )


    w_FlagSet = .false.

    if ( FlagBnds ) then

       if ( aw_InWavelenBnds(1,1) > aw_InWavelenBnds(1,2) ) then
          allocate( w_Tmp( NData ) )
          !
          do l = 1, NData
             w_Tmp(l) = aw_InWavelenBnds(1,l)
          end do
          do l = 1, NData
             aw_InWavelenBnds(1,l) = w_Tmp(NData-l+1)
          end do
          do l = 1, NData
             w_Tmp(l) = aw_InWavelenBnds(2,l)
          end do
          do l = 1, NData
             aw_InWavelenBnds(2,l) = w_Tmp(NData-l+1)
          end do
          !
          do n = 1, NPtcl
             w_Tmp = wa_InQExt(:,n)
             do l = 1, NData
                wa_InQExt(l,n) = w_Tmp(NData-l+1)
             end do
             w_Tmp = wa_InSSA(:,n)
             do l = 1, NData
                wa_InSSA(l,n) = w_Tmp(NData-l+1)
             end do
             w_Tmp = wa_InAF(:,n)
             do l = 1, NData
                wa_InAF(l,n) = w_Tmp(NData-l+1)
             end do
          end do
          !
          deallocate( w_Tmp )
       end if

       do l = 1, NWavelen
          if ( w_Wavelen(l) <= aw_InWavelenBnds(1,1) ) then
             m = 1
             w_QExt(l) = wa_InQExt(m,iParticle)
             w_SSA (l) = wa_InSSA (m,iParticle)
             w_AF  (l) = wa_InAF  (m,iParticle)
             w_FlagSet(l) = .true.
          else if ( aw_InWavelenBnds(2,NData) <= w_Wavelen(l) ) then
             m = NData
             w_QExt(l) = wa_InQExt(m,iParticle)
             w_SSA (l) = wa_InSSA (m,iParticle)
             w_AF  (l) = wa_InAF  (m,iParticle)
             w_FlagSet(l) = .true.
          else
             do m = 1, NData
                if ( ( aw_InWavelenBnds(1,m) <  w_Wavelen(l)          ) .and. &
                     ( w_Wavelen(l)          <= aw_InWavelenBnds(2,m) ) ) exit
             end do
             w_QExt(l) = wa_InQExt(m,iParticle)
             w_SSA (l) = wa_InSSA (m,iParticle)
             w_AF  (l) = wa_InAF  (m,iParticle)
             w_FlagSet(l) = .true.
          end if
       end do

    else

       if ( w_InWavelen(1) > w_InWavelen(2) ) then
          allocate( w_Tmp( NData ) )
          !
          w_Tmp = w_InWavelen(:)
          do l = 1, NData
             w_InWavelen(l) = w_Tmp(NData-l+1)
          end do
          !
          do n = 1, NPtcl
             w_Tmp = wa_InQExt(:,n)
             do l = 1, NData
                wa_InQExt(l,n) = w_Tmp(NData-l+1)
             end do
             w_Tmp = wa_InSSA(:,n)
             do l = 1, NData
                wa_InSSA(l,n) = w_Tmp(NData-l+1)
             end do
             w_Tmp = wa_InAF(:,n)
             do l = 1, NData
                wa_InAF(l,n) = w_Tmp(NData-l+1)
             end do
          end do
          !
          deallocate( w_Tmp )
       end if

       do l = 1, NWavelen
          if ( w_Wavelen(l) <= w_InWavelen(1) ) then
             m = 1
             w_QExt(l) = wa_InQExt(m,iParticle)
             w_SSA (l) = wa_InSSA (m,iParticle)
             w_AF  (l) = wa_InAF  (m,iParticle)
             w_FlagSet(l) = .true.
          else if ( w_InWavelen(NData) <= w_Wavelen(l) ) then
             m = NData
             w_QExt(l) = wa_InQExt(m,iParticle)
             w_SSA (l) = wa_InSSA (m,iParticle)
             w_AF  (l) = wa_InAF  (m,iParticle)
             w_FlagSet(l) = .true.
          else
             do m = 1+1, NData
                if ( w_Wavelen(l) <= w_InWavelen(m) ) exit
             end do
             w_QExt(l) = &
                  &   ( wa_InQExt  (m,iParticle) - wa_InQExt  (m-1,iParticle) ) &
                  & / ( w_InWavelen(m)           - w_InWavelen(m-1)           ) &
                  & * ( w_Wavelen  (l)           - w_InWavelen(m-1)           ) &
                  & + wa_InQExt(m-1,iParticle)
             w_SSA (l) = &
                  &   ( wa_InSSA   (m,iParticle) - wa_InSSA   (m-1,iParticle) ) &
                  & / ( w_InWavelen(m)           - w_InWavelen(m-1)           ) &
                  & * ( w_Wavelen  (l)           - w_InWavelen(m-1)           ) &
                  & + wa_InSSA (m-1,iParticle)
             w_AF  (l) = &
                  &   ( wa_InAF    (m,iParticle) - wa_InAF    (m-1,iParticle) ) &
                  & / ( w_InWavelen(m)           - w_InWavelen(m-1)           ) &
                  & * ( w_Wavelen  (l)           - w_InWavelen(m-1)           ) &
                  & + wa_InAF  (m-1,iParticle)
             w_FlagSet(l) = .true.
          end if
       end do

    end if


    ! Check set values
    do l = 1, NWavelen
       if ( .not. w_FlagSet(l) ) then
          write( 6, * ) "ERROR : Unset particle's optical property at l = ", l, w_FlagSet(l)
          if ( FlagBnds ) then
             write( 6, * ) "ERROR : ", aw_InWavelenBnds(1,l), aw_InWavelenBnds(2,l)
          else
             write( 6, * ) "ERROR : ", w_InWavelen(l)
          end if
          stop
       end if
    end do


    if ( FlagBnds ) then
       deallocate( aw_InWavelenBnds )
    else
       deallocate( w_InWavelen )
    end if
    deallocate( w_FlagSet    )
    deallocate( wa_InQExt    )
    deallocate( wa_InSSA     )
    deallocate( wa_InAF      )


  end subroutine ParticleECCalcSetOptProp

  !----------------------------------------------------------------------------

end module particle_ec_calc
